import pymysql
from threading import Thread, Lock, Condition, Timer


class Error(Exception):
    """Base class for all pool-related exceptions"""
    pass


class TimeoutError(Error):
    """The operation exceeded the given deadline"""
    pass


def resize_pool(interval=600.0, stati_num=3, multiple=4, counter=0,
                accumulation=0, pool=None):
    """Resize connection pool in cycle"""
    assert (pool is not None)
    with pool.lock:
        if counter >= stati_num:
            avg = accumulation / stati_num
            if avg >= multiple:
                num = len(pool.unuse_list) - pool.min_size
                for _ in range(num):
                    c = pool.unuse_list.pop()
                    c.close()
            counter = 0
            accumulation = 0
        else:
            accumulation += (len(pool.unuse_list) + 1) / (len(pool.inuse_list) + 1)
            counter += 1
    t = Timer(interval, resize_pool, args=(interval, stati_num, multiple,
                                           counter, accumulation, pool))
    t.start()


class Pool(object):
    """
    Connection pool for pymysql.

    The initialization parameters are as follows:
    :param host: Host of MySQL server
    :param port: Port of MySQL server
    :param user: User of MySQL server
    :param password: Password of MySQL server
    :param db: Database of MySQL server
    :param charset: Charset of MySQL server
    :param cursorclass: Class of MySQL Cursor
    :param autocommit: auto commit mode
    :param min_size: Minimum size of connection pool
    :param max_size: Maximum size of connection pool
    :param timeout: Watting time in the multi-thread environment
    :param interval: Statistical cycle time
    :param stati_mun: Statistical frequency
    :param multiple: Regulation standard
    :param counter: Counter
    :param accumulation: Statiscal result
    """

    def __init__(self,
                 host="localhost",
                 port=3306,
                 user=None,
                 password=None,
                 db=None,
                 charset="utf8",
                 cursorclass=pymysql.cursors.DictCursor,
                 autocommit=False,
                 min_size=1,
                 max_size=3,
                 timeout=10.0,
                 interval=600.0,
                 stati_num=3,
                 multiple=4,
                 counter=0,
                 accumulation=0):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.db = db
        self.charset = charset
        self.cursorclass = cursorclass
        self.autocommit = autocommit

        self.min_size = min_size
        self.max_size = max_size
        self.current_size = 0
        self.timeout = timeout

        self.unuse_list = set()
        self.inuse_list = set()
        self.lock = Lock()
        self.cond = Condition(self.lock)

        self.interval = interval
        self.stati_num = stati_num
        self.multiple = multiple
        self.counter = 0
        self.accumulation = 0

    def create_conn(self):
        """Create mysql connection by pymysql and to add unuse_list"""
        c = pymysql.connect(
            host=self.host,
            port=self.port,
            user=self.user,
            password=self.password,
            db=self.db,
            charset=self.charset,
            cursorclass=self.cursorclass,
            autocommit=self.autocommit
        )
        self.unuse_list.add(c)

    def _start(self):
        """Start thread for resize pool"""
        t = Thread(target=resize_pool, args=(self.interval, self.stati_num,
                                             self.multiple, self.counter,
                                             self.accumulation, self))
        t.start()

    def _init_pool(self):
        """Initial minimum size of pool"""
        assert (self.min_size <= self.max_size)
        for _ in range(self.min_size):
            self.create_conn()

    def init(self):
        self._init_pool()
        self._start()

    def _wait(self):
        """Waiting condition"""
        return len(self.unuse_list) > 0

    def get_conn(self):
        with self.cond:
            # Lack of resources and wait
            if len(self.unuse_list) <= 0 and \
                    self.current_size >= self.max_size:
                # note: TimeoutError mean release operation exception
                # or max_size much less than concurrence
                self.cond.wait_for(self._wait, self.timeout)
                if len(self.unuse_list) <= 0:
                    raise TimeoutError
            # Lack of resources but can created
            if len(self.unuse_list) <= 0 and \
                    self.current_size < self.max_size:
                self.create_conn()

            self.current_size += 1
            c = self.unuse_list.pop()
            self.inuse_list.add(c)
            return c

    def release(self, c):
        """Release connection from inuse_list to unuse_list"""
        with self.cond:
            self.current_size -= 1
            self.inuse_list.remove(c)
            self.unuse_list.add(c)
            self.cond.notify_all()

    def destroy(self):
        """Destroy pool"""
        for _ in range(len(self.unuse_list)):
            c = self.unuse_list.pop()
            c.close()
        for _ in range(len(self.inuse_list)):
            c = self.inuse_list.pop()
            c.close()



import unittest
from unittest.mock import patch
from pymysqlpool import pool
from threading import Thread


class MockMySQLConnection(object):

    def __init__(self, host='localhost', port=3306, user='root',
                 password='None', db='None', charset='utf8'):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.db = db
        self.charset = charset

    def close(self):
        return 'Mock pymysql close method.'


# mock method of pool
def mock_create_conn(self):
    c = MockMySQLConnection()
    self.unuse_list.add(c)


def mock__init_pool(self):
    assert (self.min_size <= self.max_size)
    for _ in range(self.min_size):
        mock_create_conn(self)


def mock__start(self):
    pass


# test case
class TestPool(unittest.TestCase):

    def setUp(self):
        self.concurrence = 1000
        # unequal pre
        self.p = pool.Pool(min_size=10, max_size=90, timeout=10.0)

    def tearDown(self):
        pass

    @patch.object(pool.Pool, 'create_conn', new=mock_create_conn)
    def test_create_conn(self):
        assert (mock_create_conn == pool.Pool.create_conn)
        self.p.create_conn()
        assert (len(self.p.unuse_list) == 1)
        assert (filter(lambda x: isinstance(x, MockMySQLConnection), self.p.unuse_list))

    @patch.object(pool.Pool, 'create_conn', new=mock_create_conn)
    @patch.object(pool.Pool, '_init_pool', new=mock__init_pool)
    @patch.object(pool.Pool, '_start', new=mock__start)
    def test_init(self):
        self.p.init()
        assert (len(self.p.unuse_list) == self.p.min_size)

    @patch.object(pool.Pool, 'create_conn', new=mock_create_conn)
    @patch.object(pool.Pool, '_init_pool', new=mock__init_pool)
    @patch.object(pool.Pool, '_start', new=mock__start)
    def test_get_conn(self):
        self.p.init()
        c = self.p.get_conn()
        assert (isinstance(c, MockMySQLConnection))
        assert (len(self.p.unuse_list) == self.p.min_size - 1)
        assert (len(self.p.inuse_list) == 1)
        assert (filter(lambda x: isinstance(x, MockMySQLConnection), self.p.unuse_list))
        assert (filter(lambda x: isinstance(x, MockMySQLConnection), self.p.inuse_list))

    @patch.object(pool.Pool, 'create_conn', new=mock_create_conn)
    @patch.object(pool.Pool, '_init_pool', new=mock__init_pool)
    @patch.object(pool.Pool, '_start', new=mock__start)
    def test_release(self):
        self.p.init()
        c = self.p.get_conn()
        self.p.release(c)
        assert (len(self.p.unuse_list) == self.p.min_size)
        assert (len(self.p.inuse_list) == 0)

    @patch.object(pool.Pool, 'create_conn', new=mock_create_conn)
    @patch.object(pool.Pool, '_init_pool', new=mock__init_pool)
    @patch.object(pool.Pool, '_start', new=mock__start)
    def test_destroy(self):
        self.p.init()
        self.p.inuse_list.add(MockMySQLConnection())
        self.p.destroy()
        assert (len(self.p.unuse_list) == 0)
        assert (len(self.p.inuse_list) == 0)

    @patch.object(pool.Pool, 'create_conn', new=mock_create_conn)
    @patch.object(pool.Pool, '_init_pool', new=mock__init_pool)
    @patch.object(pool.Pool, '_start', new=mock__start)
    def pure_get_and_release(self):
        c = self.p.get_conn()
        self.p.release(c)

    @patch.object(pool.Pool, 'create_conn', new=mock_create_conn)
    @patch.object(pool.Pool, '_init_pool', new=mock__init_pool)
    @patch.object(pool.Pool, '_start', new=mock__start)
    def test_get_conn_by_concurrence(self):
        self.p.init()
        for i in range(self.concurrence):
            t = Thread(target=self.pure_get_and_release)
            t.start()
        assert (len(self.p.unuse_list) == self.p.min_size)
        assert (len(self.p.inuse_list) == 0)
        assert (filter(lambda x: isinstance(x, MockMySQLConnection), self.p.unuse_list))


if __name__ == '__main__':
    unittest.main(verbosity=2)
